"""Contain information about the project."""
__package_name__ = 'opensignals'
__version__ = '0.0.1-rc7'

__description__ = 'Open Signals'

__author__ = 'Jordi Villar'
__author_email__ = 'jrdi.villar@gmail.com'

__license__ = 'Apache License'
